// bignum2.cpp : 
// This project represents large integer numbers.
// Each number is represented as a list of integers rather
// than by using a single integer. This allows us to represent very large
// numbers. It also allows very accurate arithmetic since the normal integer
// or floating point arithmetic has roundoff problems and is limited in size.
// Arithmetic operations are implemented to simulate the way people do 
// arithmetic when using pencil and paper. 
// The output and arithmetic operations use operator overloading.

#include "stdafx.h"
#include <list>
#include <iostream>

using namespace std;

// exception class used if non-digit passed to constructor
// It saves the bad character to be displayed later
class baddigit {
	public:
	   baddigit(char x) { mychar = x;}
	   char get_char() { return mychar;}
	   friend ostream & operator<<(ostream & os, baddigit & b) {
		   os << "bad char was :" << b.get_char();
		   return os;
	   }
	private:
	   char mychar;
}; // baddigit

// this class represents the whole number. Each digit is stored as a
// seperate element in an STL list object.
// The public methods provide a layer between the built in list operations
// and users of the bignum.
// Instances of the bignum class are initialized by passing in a 
// character string of integers, for example, "12345". 
// The constructor uses another method
// to break the string into individual digits which are stored in the list.
// The digits are stored in reverse order in the list to make the arithmetic
// a little easier.
class bignum {
public:
   bignum() {}
   bignum(char *str) {this->read_string(str);}  // constructor
   int how_big() { return(L.size());}    // returns number of digits
   int is_empty() { return(L.size() == 0);}  // if no digits, it is empty
   list<int> get_L() { return L;}  // returns the list object for use elsewhere
   list<int>::iterator get_begin() { return L.begin();}
   list<int>::iterator get_end() { return L.end();}
   void add_digit(int n) {L.push_front(n);}  // add a digit to the front of the list
   void read_string(char * str);  // convert char array to list
   void push_digit(int n) { L.push_back(n);} // add digit to end of list

   friend ostream& operator<<(const ostream& os, const bignum& num); // print number
   friend bignum * operator+(bignum & left,bignum & right); // add two bignums
protected:
   list<int> L;  // STL list object made up of integers
};

// takes each letter from the string of digits and converts to
// the equivalent integer. This integer is put on the front of 
// the list so "12345" is stored as 5->4->3->2->1.
// The conversion from the letter '2' to the number 2 is done by subtracting 
// the ASCII values of '2' and '0'. The ASCII value of '2' is 32 and the ASCII
// value of '0' is 30 so '2' - '0' == 2.
// Note that we are using the character zero ('0') not the NUL character ('\0').
void
bignum::read_string(char * str)
{
   for(int i=0;i<strlen(str);i++) {
      if(isdigit(str[i])) {  // make sure this is a digit
      	add_digit(str[i]-'0');
      }
      else
      	throw baddigit(str[i]);
   } // for
} // read_string 

// walk the bignum, printing each nodes digit 
// overloaded insertor to print the values in the bignum.
// This allows us to use bignums in output streams without having to
// call methods. This could be used like this
// bignum b("12345");
// cout << b << endl;
// Since the digits are stored in reverse order in the list, we start the
// iterator at the end and work towards the front.
ostream & 
operator<<(ostream& os, bignum& num)
{
     list<int>::iterator ptr = num.get_end();
     while (ptr != num.get_begin()) {
	 ptr--;
         os << *ptr;  // retrieve and print the integer ptr points at
     }
     return os;   // always return the ostream argument
} // operator<<

// This replaces the + operator when used with bignums so if we have
// bignum a("12345"), b("54321"), *c;
// c = a + b;
// The previous line turns into a function call like this
// c = operator+(a,b);
//
// This code simulates (sort of) the way people add numbers.
// It starts with the rightmost pair of digits from the numbers,
// adds them and stores the resulting digit in the answer. If the result
// is greater than 10, it save the ones digit and carries the one to the
// next place.
// To avoid some if statements, we initialize the carry to 0 and always add it.
// Kind of as if there were some extra zeros on the right.
// Also, if one number is shorter than the other, we continue adding as if
// we had added zeros to the left to make the numbers the same length.
// Doing this make the code a little simpler with only a small performance hit.
// Since we stored the digits in reverse order we can add the digits starting
// at the front of the list rather than the end.
// Also, this doesn't handle negative numbers.
bignum *
operator+(bignum& left,bignum& right) 
{
   int carry=0,one,two,three;
   list<int>::iterator  lptr ,rptr;
   bignum *result=new bignum();

   // start the iterators at the beginning of the lists
   rptr=right.get_begin(); lptr=left.get_begin();

   while(true) { // infinite loop
      // for each bignum, check if there are digits left. If there
      // are not, set the variable to -1. This will tell us later
      // whether to add it into the result or not.
      one = (rptr != right.get_end()) ? *rptr : -1;
      two = (lptr != left.get_end()) ? *lptr : -1;

      if((one == -1) && (two == -1)) { // both done
	 // even though all the digits on both bignums have been added, 
	 // there may be a carry from  the last operation
         if(carry > 0)
            result->push_digit(carry);
         return result;
      }
      // we do this here because then we don't have to check for the
      // usual case nor do we need a nested if-then-else thing below.
      // We trade a possible wasted operation for less code and a little
      // clearer structure
      three = one + two + carry;
     
      if(one == -1)  // copy rest of two
            three = two + carry;  // pretend that right was padded with zeroes
      if(two == -1)  // copy rest of one
            three = one + carry;  // pretend that left was padded with zeroes
      carry = 0;  // reset the carry 

      if(three >=  10) {
        carry=1;
        three -= 10;
      }
      result->push_digit(three); // save the result digit of this pass

      // only increment the pointers if there are still digits left
      if(lptr != left.get_L().end()) lptr++; 
      if(rptr != right.get_L().end()) rptr++;
   } // while
} // operator+

int 
main(int argc, char* argv[])
{
   bignum one,two,*three;

   try {
      one.read_string("12345"); two.read_string("54321");

      cout << "bignum one has " << one.how_big() << " elements" << endl;
      cout << one << endl;

      cout << "bignum two has " << two.how_big() << " elements" << endl;
      cout << two << endl;

      three=one + two; //two.add(one);
      cout << "here is the final bignum three " << *three << endl;
      cout << " now test the exception" << endl;
      bignum bad("123bad");
   }
   catch(baddigit b) {
	   cerr << b << endl;
   }

   return 0;
} // main
